"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const is_plain_object_1 = __importDefault(require("is-plain-object"));
const createWorkers_1 = require("./createWorkers");
const SharedData_1 = require("./SharedData");
const SmartSubscriber_1 = require("./SmartSubscriber");
const types_1 = require("./types");
const sharedData = new SharedData_1.SharedData();
const { createNewValue, computeUpdates } = createWorkers_1.createWorkers(sharedData);
class LemonStore {
    constructor(initialState, name = 'Unknown') {
        this.name = name;
        this.id = Symbol(this.name);
        this.isRemoved = false;
        this.getState = () => {
            return this.proxy;
        };
        this.subscribe = (fn) => {
            this.subscribers.push(fn);
            return () => {
                this.subscribers = this.subscribers.filter(sub => sub !== fn);
            };
        };
        this.smartSubscribe = (fn) => {
            return this.smartSubscriber.smartSubscribe(fn);
        };
        this.setState = (diff) => {
            if (!is_plain_object_1.default(diff)) {
                sharedData.throwError(TypeError, `New state must be plain Object (in ${this.name})`);
            }
            if (!sharedData.canSetState()) {
                sharedData.throwError(Error, `Can't modify any state when values is computing (in ${this.name})`);
            }
            const newStateEntries = Object.entries(diff);
            const firstLevelDepsArr = new Array();
            let isObjectsEquals = true;
            newStateEntries.forEach(([key, value]) => {
                if (this.stateValues.has(key)) {
                    const valueInfo = sharedData.getValueInfoByStateProp(this.stateValues, key);
                    if (valueInfo.type === types_1.ValueType.COMPUTED) {
                        sharedData.throwError(TypeError, `Can't modify '${key}' property (in ${this.name}), this is computed value.`);
                    }
                    if (!Object.is(valueInfo.value, value)) {
                        valueInfo.value = value;
                        valueInfo.cachedValue = value;
                        sharedData.registerChangedValue(valueInfo.id);
                        isObjectsEquals = false;
                        firstLevelDepsArr.push(...Array.from(valueInfo.invoking));
                    }
                }
                else {
                    isObjectsEquals = false;
                    this.stateValues.set(key, createNewValue({
                        storeId: this.id,
                        name: this.name,
                        proxy: this.proxy
                    }, value, key, true));
                }
            });
            if (!isObjectsEquals) {
                computeUpdates(new Set(firstLevelDepsArr));
            }
        };
        if (!is_plain_object_1.default(initialState)) {
            sharedData.throwError(TypeError, `InitialState (in ${name}) must be plain Object`);
        }
        const updateProxyLink = () => {
            this.proxy = Object.create(Object.getPrototypeOf(this.proxy), Object.getOwnPropertyDescriptors(this.proxy));
        };
        sharedData.registerStore(this.id, this, () => this.subscribers, updateProxyLink);
        this.proxy = {};
        const initStateKeys = Object.keys(initialState);
        this.stateValues = new Map();
        initStateKeys.forEach((key) => {
            const value = initialState[key];
            this.stateValues.set(key, createNewValue({
                storeId: this.id,
                name: this.name,
                proxy: this.proxy
            }, value, key));
        });
        initStateKeys.forEach((key) => this.proxy[key]);
        this.subscribers = [];
        this.smartSubscriber = new SmartSubscriber_1.SmartSubscriber(this);
    }
    remove() {
        if (!sharedData.canSetState()) {
            sharedData.throwError(Error, `Can't remove store in computed value (in ${this.name})`);
        }
        this.subscribers = [];
        const throwAccessError = () => {
            return sharedData.throwError(ReferenceError, `Store is removed! (in ${this.name})`);
        };
        this.proxy = {};
        this.setState = throwAccessError;
        this.subscribe = throwAccessError;
        this.smartSubscribe = throwAccessError;
        this.smartSubscriber.remove();
        const values = Array.from(this.stateValues).map(([_, value]) => value);
        this.name = `${this.name}[removed]`;
        sharedData.removeStoreValues(new Set(values));
        sharedData.removeStoreProxyLinkUpdater(this.id);
        sharedData.removeStoreSubscribers(this.id);
        this.stateValues.clear();
        this.isRemoved = true;
    }
}
exports.LemonStore = LemonStore;
